@@ -1,11 +1,13 @@
 package Algorithm::Diff;
 use strict;
 use vars qw($VERSION @EXPORT_OK @ISA @EXPORT);
+use integer;		# see below in _replaceNextLargerWith() for mod to make
+					# if you don't use this
 require Exporter;
 @ISA = qw(Exporter);
 @EXPORT = qw();
 @EXPORT_OK = qw(LCS diff traverse_sequences);
-$VERSION = sprintf('%d.%02d', (q$Revision: 1.6 $ =~ /\d+/g));
+$VERSION = sprintf('%d.%02d', (q$Revision: 1.11 $ =~ /\d+/g));
 
 # McIlroy-Hunt diff algorithm
 # Adapted from the Smalltalk code of Mario I. Wolczko, <mario@wolczko.com>
@@ -178,30 +180,27 @@ routine.
 C<traverse_sequences> is the most general facility provided by this
 module; C<diff> and C<LCS> are implemented as calls to it.
 
-Imagine that there are two arrows.  Arrow A points to an element of
-sequence A, and arrow B points to an element of the sequence B.
-Initially, the arrows point to the first elements of the respective
-sequences.  C<traverse_sequences> will advance the arrows through the
-sequences one element at a time, calling an appropriate user-specified
-callback function before each advance.  It willadvance the arrows in
-such a way that if there are equal elements C<$A[$i]> and C<$B[$j]>
-which are equal and which are part of the LCS, there will be some
-moment during the execution of C<traverse_sequences> when arrow A is
-pointing to C<$A[$i]> and arrow B is pointing to C<$B[$j]>.  When this
-happens, C<traverse_sequences> will call the C<MATCH> callback
-function and then it will advance both arrows. 
+Imagine that there are two arrows.  Arrow A points to an element of sequence A,
+and arrow B points to an element of the sequence B.  Initially, the arrows
+point to the first elements of the respective sequences.  C<traverse_sequences>
+will advance the arrows through the sequences one element at a time, calling an
+appropriate user-specified callback function before each advance.  It
+willadvance the arrows in such a way that if there are equal elements C<$A[$i]>
+and C<$B[$j]> which are equal and which are part of the LCS, there will be
+some moment during the execution of C<traverse_sequences> when arrow A is
+pointing to C<$A[$i]> and arrow B is pointing to C<$B[$j]>.  When this happens,
+C<traverse_sequences> will call the C<MATCH> callback function and then it will
+advance both arrows. 
 
-Otherwise, one of the arrows is pointing to an element of its sequence
-that is not part of the LCS.  C<traverse_sequences> will advance that
-arrow and will call the C<DISCARD_A> or the C<DISCARD_B> callback,
-depending on which arrow it advanced.  If both arrows point to
-elements that are not part of the LCS, then C<traverse_sequences> will
-advance one of them and call the appropriate callback, but it is not
-specified which it will call.
+Otherwise, one of the arrows is pointing to an element of its sequence that is
+not part of the LCS.  C<traverse_sequences> will advance that arrow and will
+call the C<DISCARD_A> or the C<DISCARD_B> callback, depending on which arrow it
+advanced.  If both arrows point to elements that are not part of the LCS, then
+C<traverse_sequences> will advance one of them and call the appropriate
+callback, but it is not specified which it will call.
 
-The arguments to C<traverse_sequences> are the two sequences to
-traverse, and a callback which specifies the callback functions, like
-this:
+The arguments to C<traverse_sequences> are the two sequences to traverse, and a
+callback which specifies the callback functions, like this:
 
   traverse_sequences( \@seq1, \@seq2,
                      { MATCH => $callback_1,
@@ -209,43 +208,41 @@ this:
                        DISCARD_B => $callback_3,
                      } );
 
-Callbacks are invoked with at least the indices of the two arrows as
-their arguments.  They are not expected to return any values.  If a
-callback is omitted from the table, it is not called.
+Callbacks for MATCH, DISCARD_A, and DISCARD_B are invoked with at least the
+indices of the two arrows as their arguments.  They are not expected to return
+any values.  If a callback is omitted from the table, it is not called.
+
+Callbacks for A_FINISHED and B_FINISHED are invoked with at least the
+corresponding index in A or B,
 
 If arrow A reaches the end of its sequence, before arrow B does,
-C<traverse_sequences> will call the C<A_FINISHED> callback when it
-advances arrow B, if there is such a function; if not it will call
-C<DISCARD_B> instead.  Similarly if arrow B finishes first.
-C<traverse_sequences> returns when both arrows are at the ends of
-their respective sequences.  It returns true on success and false on
-failure.  At present there is no way to fail.
+C<traverse_sequences> will call the C<A_FINISHED> callback when it advances
+arrow B, if there is such a function; if not it will call C<DISCARD_B> instead.
+Similarly if arrow B finishes first.  C<traverse_sequences> returns when both
+arrows are at the ends of their respective sequences.  It returns true on
+success and false on failure.  At present there is no way to fail.
 
-C<traverse_sequences> may be passed an optional fourth parameter; this
-is a CODE reference to a key generation function.  See L</KEY
-GENERATION FUNCTIONS>.
+C<traverse_sequences> may be passed an optional fourth parameter; this is a
+CODE reference to a key generation function.  See L</KEY GENERATION FUNCTIONS>.
 
-Additional parameters, if any, will be passed to the key generation
-function.
+Additional parameters, if any, will be passed to the key generation function.
 
 =head1 KEY GENERATION FUNCTIONS
 
 C<diff>, C<LCS>, and C<traverse_sequences> accept an optional last parameter.
 This is a CODE reference to a key generating (hashing) function that should
-return a string that uniquely identifies a given element.
-It should be the case that if two elements are to be considered equal,
-their keys should be the same (and the other way around).
-If no key generation function is provided, the key will be the
-element as a string.
+return a string that uniquely identifies a given element.  It should be the
+case that if two elements are to be considered equal, their keys should be the
+same (and the other way around).  If no key generation function is provided,
+the key will be the element as a string.
 
 By default, comparisons will use "eq" and elements will be turned into keys
 using the default stringizing operator '""'.
 
-Where this is important is when you're comparing something other than
-strings. If it is the case that you have multiple different objects 
-that should be considered to be equal, you should supply a key
-generation function. Otherwise, you have to make sure that your arrays
-contain unique references.
+Where this is important is when you're comparing something other than strings.
+If it is the case that you have multiple different objects that should be
+considered to be equal, you should supply a key generation function. Otherwise,
+you have to make sure that your arrays contain unique references.
 
 For instance, consider this example:
 
@@ -370,9 +367,8 @@ sub _withPositionsOfInInterval
 
 sub _replaceNextLargerWith
 {
-	my $array = shift;	# array ref
-	my $aValue = shift;	# numeric
-	my $high = $#$array;
+	my ( $array, $aValue, $high ) = @_;
+	$high ||= $#$array;
 
 	# off the end?
 	if ( $high == -1 || $aValue > $array->[ -1 ] )
@@ -384,10 +380,18 @@ sub _replaceNextLargerWith
 	# binary search for insertion point...
 	my $low = 0;
 	my $index;
+	my $found;
 	while ( $low <= $high )
 	{
 		$index = ( $high + $low ) / 2;
-		if ( $aValue > $array->[ $index ] )
+#		$index = int(( $high + $low ) / 2);		# without 'use integer'
+		$found = $array->[ $index ];
+
+		if ( $aValue == $found )
+		{
+			return undef;
+		}
+		elsif ( $aValue > $found )
 		{
 			$low = $index + 1;
 		}
@@ -396,9 +400,8 @@ sub _replaceNextLargerWith
 			$high = $index - 1;
 		}
 	}
-	# now insertion point is in $low.
-	return undef if $array->[ $low ] == $aValue;	# same?
 
+	# now insertion point is in $low.
 	$array->[ $low ] = $aValue;		# overwrite next larger
 	return $low;
 }
@@ -406,8 +409,8 @@ sub _replaceNextLargerWith
 # This method computes the longest common subsequence in $a and $b.
 
 # Result is array or ref, whose contents is such that
-# 	$a->[ $i ] = $b->[ $result[ $i ] ]
-# foreach $i in ( 0..scalar( @result ) if $result[ $i ] is defined.
+# 	$a->[ $i ] == $b->[ $result[ $i ] ]
+# foreach $i in ( 0 .. $#result ) if $result[ $i ] is defined.
 
 # An additional argument may be passed; this is a hash or key generating
 # function that should return a string that uniquely identifies the given
@@ -424,20 +427,32 @@ sub _longestCommonSubsequence
 {
 	my $a = shift;	# array ref
 	my $b = shift;	# array ref
-	my $keyGen = shift || sub { $_[0] };
-	my $compare = sub { &$keyGen( $_[0], @_ ) eq &$keyGen( $_[1], @_ ) };
+	my $keyGen = shift;	# code ref
+	my $compare;	# code ref
 
-	my( $aStart, $aFinish, $bStart, $bFinish ) = ( 0, $#$a, 0, $#$b );
-	my $matchVector = [];
+	# set up code refs
+	# Note that these are optimized.
+	if ( !defined( $keyGen ) )	# optimize for strings
+	{
+		$keyGen = sub { $_[0] };
+		$compare = sub { my ($a, $b) = @_; $a eq $b };
+	}
+	else
+	{
+		$compare = sub {
+			my $a = shift; my $b = shift;
+			&$keyGen( $a, @_ ) eq &$keyGen( $b, @_ )
+		};
+	}
+
+	my ($aStart, $aFinish, $bStart, $bFinish, $matchVector) = (0, $#$a, 0, $#$b, []);
 
 	# First we prune off any common elements at the beginning
 	while ( $aStart <= $aFinish
 		and $bStart <= $bFinish
 		and &$compare( $a->[ $aStart ], $b->[ $bStart ], @_ ) )
 	{
-		$matchVector->[ $aStart ] = $bStart;
-		$aStart++;
-		$bStart++;
+		$matchVector->[ $aStart++ ] = $bStart++;
 	}
 
 	# now the end
@@ -445,21 +460,18 @@ sub _longestCommonSubsequence
 		and $bStart <= $bFinish
 		and &$compare( $a->[ $aFinish ], $b->[ $bFinish ], @_ ) )
 	{
-		$matchVector->[ $aFinish ] = $bFinish;
-		$aFinish--;
-		$bFinish--;
+		$matchVector->[ $aFinish-- ] = $bFinish--;
 	}
 
 	# Now compute the equivalence classes of positions of elements
-	my $bMatches = _withPositionsOfInInterval( $b, $bStart, $bFinish,
-		$keyGen, @_ );
+	my $bMatches = _withPositionsOfInInterval( $b, $bStart, $bFinish, $keyGen, @_ );
 	my $thresh = [];
 	my $links = [];
 
 	my ( $i, $ai, $j, $k );
 	for ( $i = $aStart; $i <= $aFinish; $i++ )
 	{
-		$ai = &$keyGen( $a->[ $i ] );
+		$ai = &$keyGen( $a->[ $i ], @_ );
 		if ( exists( $bMatches->{ $ai } ) )
 		{
 			$k = 0;
@@ -470,12 +482,13 @@ sub _longestCommonSubsequence
 					and $thresh->[ $k ] > $j
 					and $thresh->[ $k - 1 ] < $j )
 				{
-					$thresh->[ $k ] = $j
+					$thresh->[ $k ] = $j;
 				}
 				else
 				{
-					$k = _replaceNextLargerWith( $thresh, $j )
+					$k = _replaceNextLargerWith( $thresh, $j, $k );
 				}
+
 				# oddly, it's faster to always test this (CPU cache?).
 				if ( defined( $k ) )
 				{
@@ -502,35 +515,51 @@ sub traverse_sequences
 	my $a = shift;	# array ref
 	my $b = shift;	# array ref
 	my $callbacks = shift || { };
-	my $compare = shift;
 	my $keyGen = shift;
 	my $matchCallback = $callbacks->{'MATCH'} || sub { };
 	my $discardACallback = $callbacks->{'DISCARD_A'} || sub { };
+	my $finishedACallback = $callbacks->{'A_FINISHED'};
 	my $discardBCallback = $callbacks->{'DISCARD_B'} || sub { };
-	my $matchVector = _longestCommonSubsequence( $a, $b,
-		$compare, $keyGen, @_ );
+	my $finishedBCallback = $callbacks->{'B_FINISHED'};
+	my $matchVector = _longestCommonSubsequence( $a, $b, $keyGen, @_ );
 	# Process all the lines in match vector
 	my $lastA = $#$a;
 	my $lastB = $#$b;
 	my $bi = 0;
 	my $ai;
+
 	for ( $ai = 0; $ai <= $#$matchVector; $ai++ )
 	{
 		my $bLine = $matchVector->[ $ai ];
-		if ( defined( $bLine ) )
+		if ( defined( $bLine ) )	# matched
 		{
 			&$discardBCallback( $ai, $bi++, @_ ) while $bi < $bLine;
-			&$matchCallback( $ai, $bi, @_ );
-			$bi++;
+			&$matchCallback( $ai, $bi++, @_ );
 		}
 		else
 		{
 			&$discardACallback( $ai, $bi, @_ );
 		}
 	}
+	# the last entry (if any) processed was a match.
 
-	&$discardACallback( $ai++, $bi, @_ ) while ( $ai <= $lastA );
-	&$discardBCallback( $ai, $bi++, @_ ) while ( $bi <= $lastB );
+	if ( defined( $finishedBCallback ) && $ai <= $lastA )
+	{
+		&$finishedBCallback( $bi, @_ );
+	}
+	else
+	{
+		&$discardACallback( $ai++, $bi, @_ ) while ( $ai <= $lastA );
+	}
+
+	if ( defined( $finishedACallback ) && $bi <= $lastB )
+	{
+		&$finishedACallback( $ai, @_ );
+	}
+	else
+	{
+		&$discardBCallback( $ai, $bi++, @_ ) while ( $bi <= $lastB );
+	}
 	return 1;
 }
 
